package net.datacrow.connector;

import net.datacrow.console.GUI;
import net.datacrow.console.windows.security.LoginDialog;
import net.datacrow.core.DcConfig;
import net.datacrow.core.DcRepository;
import net.datacrow.core.console.IMasterView;
import net.datacrow.core.console.IView;
import net.datacrow.core.data.DataFilters;
import net.datacrow.core.data.DcIconCache;
import net.datacrow.core.drivemanager.DriveManager;
import net.datacrow.core.enhancers.ValueEnhancers;
import net.datacrow.core.filerenamer.FilePatterns;
import net.datacrow.core.modules.DcModules;
import net.datacrow.core.modules.upgrade.ModuleUpgrade;
import net.datacrow.core.resources.DcResources;
import net.datacrow.core.security.SecuredUser;
import net.datacrow.core.services.Servers;
import net.datacrow.core.utilities.ServerUtilities;
import net.datacrow.core.utilities.SystemMonitor;
import net.datacrow.server.LocalServerConnector;
import net.datacrow.server.db.DatabaseManager;
import net.datacrow.server.security.SecurityCenter;
import net.datacrow.settings.DcSettings;
import net.datacrow.tabs.Tabs;

import org.apache.log4j.Logger;

public class DirectConnector extends LocalServerConnector {
	
	private static Logger logger = Logger.getLogger(DirectConnector.class.getName());

	public DirectConnector() {
		super();
	}
	
    @Override
	public String getUsername() {
		return super.getUsername() == null ? "SA" : super.getUsername();
	}
    
    @Override
    public void displayError(String msg) {
    	GUI.getInstance().displayErrorMessage(msg);
    }
    
    @Override
    public void displayMessage(String msg) {
    	GUI.getInstance().displayMessage(msg);
    }
    
    @Override
    public void applySettings() {
        GUI.getInstance().getMainFrame().applySettings(false);
    }
	
	@Override
	public void initialize() {
	    try {	        
            Servers.getInstance();

	        new ModuleUpgrade().upgrade();
			DcModules.load();
	        
	        SecurityCenter.getInstance().initialize();
	            
	        login(getUsername(), getPassword());

	        DcSettings.set(DcRepository.Settings.stConnectionString, "dc");
	        String db = DcConfig.getInstance().getDatabaseName();
	        if (!ServerUtilities.isEmpty(db))
	            DcSettings.set(DcRepository.Settings.stConnectionString, db);

	        DcSettings.initialize();
	        
	        DatabaseManager.getInstance().initialize();
	        DcModules.loadDefaultModuleData();
	        
	        DataFilters.load();
	        FilePatterns.load();
	        ValueEnhancers.initialize();
	        
	        SystemMonitor monitor = new SystemMonitor();
	        monitor.start();
	        
	    } catch (Exception e) {
	        logger.error(e, e);
	    }
	}
	
    @Override
    public SecuredUser login(String user, String pw) {
        
        String username = getUsername();
        String password = getPassword();
        
        SecuredUser su = null;
        
        // SA is the default, so ignore this for this routine (user left the user name empty)
        if ("SA".equals(username) && password == null)
            username = null;
        
        if (username == null) {
            su = super.login("SA", "");
            if (su == null) {
                boolean success = false;
                int retry = 0;
                
                GUI.getInstance().showSplashScreen(false);
                while (!success && retry < 3) {
                                       
                    LoginDialog dlg = new LoginDialog();
                    GUI.getInstance().openDialogNativeModal(dlg);
                    if (dlg.isCanceled()) break;
                    
                    su = super.login(dlg.getLoginName(), dlg.getPassword());
                    
                    success = su != null;
                }
                
                if (!success) 
                    System.exit(0);
                else
                    GUI.getInstance().showSplashScreen(true);
            }
        } else {
            // use the blunt message
            su = super.login(username, password);
        }
        
        return su;
    }

    @Override
    public void shutdown(boolean saveChanges) {
        DcConfig dcc = DcConfig.getInstance();
        dcc.getClientSettings().save();
        dcc.getConnector().close();

        DriveManager.getInstance().stopScanners();
        DriveManager.getInstance().stopDrivePoller();
        DriveManager.getInstance().stopFileSynchronizer();
        
        logger.info(DcResources.getText("msgApplicationStops"));
        
        if (saveChanges) {
            boolean unsavedChanges = false;
            for (IMasterView mv : GUI.getInstance().getViews()) {
                mv.saveSettings();
                for (IView view : mv.getViews()) {
                    if (!view.isChangesSaved())
                        unsavedChanges = true;
                }
            }
            
            if (unsavedChanges && !GUI.getInstance().displayQuestion("msgCancelExitAndSave"))
                return;
        }
        
        Tabs.getInstance().save();
        
        DcIconCache.getInstance().deleteIcons();
        
        if (saveChanges) {
            DataFilters.save();
            FilePatterns.save();
            DcSettings.save();
            DcModules.save();
        }
        
        DatabaseManager.getInstance().closeDatabases(false);
        
        GUI.getInstance().getMainFrame().setVisible(false);

        System.exit(0);
    }
}
